
/*
 * Baby Monitoring System with IoT Integration
 * Description: 

    we developed an advanced baby monitoring system that combines various IoT technologies to ensure the comfort and safety of a baby while providing convenience for the parents.
    This system is packed with several features that work seamlessly together to monitor the baby's environment and activities.
    
    Cry Detection and Notification: The system includes a sound sensor that continuously monitors the noise levels in the baby's room. When the baby cries, 
    the sensor detects the elevated sound levels and, using the Make automation platform, immediately sends a text message to the mother's phone. This ensures that the parent is promptly alerted, 
    even if they are in another room or busy with other tasks. The sound levels are also displayed in real-time on the Blynk app as a graph, allowing the parent to keep an eye on the noise environment in the room.
    
    Diaper Wetness Alert: To help the parent know when the baby needs a diaper change, the system is equipped with a moisture sensor that detects when the diaper is wet.
    Upon detection, the system sends a text message to the mother's phone through Make, notifying her of the need for a diaper change. Additionally, a red LED indicator lights up in the Blynk app, 
    clearly signaling that the moisture level is high and attention is needed.
    
    Lullaby Playback: To soothe the baby, the system allows the parent to play a lullaby remotely. By pressing a button in the Blynk app, a lullaby is played through a speaker connected to the system. 
    This feature is particularly useful for calming the baby without needing to be physically present.
    
    Crib Movement Control: The system includes a servo motor connected to a joystick, which controls the rocking of the baby's crib. When the joystick button is pressed, the servo motor is activated,
    moving the crib gently with the help of a thread connected to it. Pressing the button again stops the motion, giving the parent full control over the crib's movement.
    
    Temperature Monitoring: To ensure that the baby's milk bottle is at the right temperature, the system is equipped with a temperature sensor.
    The temperature is continuously monitored and displayed on the Blynk app, giving the parent real-time information and peace of mind that the baby's milk is neither too hot nor too cold.
    
    Night Lamp Control: The system also includes a white LED that functions as a night lamp. The parent can easily turn the night lamp on or off via a button in the Blynk app, 
    providing gentle illumination in the baby's room as needed.
    
    This comprehensive baby monitoring system not only provides essential information and alerts to the parent but also offers remote control over the baby's environment,
    ensuring that the baby is always safe, comfortable, and well-cared for.

    
 * The Circuit 
 * ESP32:
      Outputs:
        *GPIO 27: Connected to the data pin of the LED.
        *GPIO 23: Connected to the servo mototr.
        
        *I2S Pins:
          *GPIO 26: Connected to the Bit Clock (BCLK) of the I2S DAC or amplifier.
          *GPIO 25: Connected to the LRC 
          *GPIO 22: Connected to the DIN
          *Speaker connected to the I2S: L+ on MAX98357 -> Positive terminal of the speaker.
                                         L- on MAX98357 -> Negative terminal of the speaker.

      Inputs: 
        *GPIO 34: Connected to the moisture sensor.
        *GPIO 4: connected to the Temperature & humidity sensor
        *GPIO 35 : connected to the sound sensor (microphone)
        
 
 *Video link: https://drive.google.com/file/d/1cVe1koyDC_3_QD48nJI53oypFKCagQ10/view?usp=sharing   
 
 *Created by: 
   Daphne Messing 322630518
   Shani White 207880394

*/


#include <Arduino.h>
#include <WiFi.h>
#include <HTTPClient.h>
#include <ESP32Servo.h>
#include <DHT.h>
#include "SPIFFS.h"
#include "AudioFileSourceSPIFFS.h"
#include "AudioFileSourceID3.h"
#include "AudioGeneratorMP3.h"
#include "AudioOutputI2S.h"

// Blynk template details for the project
#define BLYNK_TEMPLATE_ID "TMPL6MIBflCH5"
#define BLYNK_TEMPLATE_NAME "Final Project"
#define BLYNK_AUTH_TOKEN "3164cQIZocDipJBfxOifMe53KFCd70-T"
#include <BlynkSimpleEsp32.h>

// WiFi credentials
const char* ssid = "Daphne";
const char* password = "negev3107";

// Webhook URLs for sending notifications
const char* serverName = "https://hook.eu2.make.com/6xfhfy45ei51hoj0cei352o3rkmtragr";
const char* waterServerName = "https://hook.eu2.make.com/37yw7teb5l2sxyrkhkum76o6r3topgu4"; // New webhook URL for water detection

// Audio components for playing MP3 files
AudioGeneratorMP3 *mp3 = nullptr;
AudioFileSourceSPIFFS *file = nullptr;
AudioOutputI2S *out = nullptr;
AudioFileSourceID3 *id3 = nullptr;
bool isPlaying = false;

// Blynk authentication token
char auth[] = BLYNK_AUTH_TOKEN;

// Defining hardware connections and states
Servo myServo;
const int soundSensorPin = 35;         // Pin where the sound sensor is connected
const int soilSensorPin = 34;          // Pin where the soil/water sensor is connected
const int joystickButtonPin = 2;       // Pin where the joystick button (SW) is connected
bool servoActive = false;              // Tracks whether the servo is active
bool lastButtonState = LOW;            // Last state of the joystick button
bool currentButtonState = LOW;         // Current state of the joystick button
int soundLevel;                        // Variable to store the sound level
int soilMoistureLevel;                 // Variable to store the soil moisture level
bool notificationSent = false;         // Flag to track if a crying notification has been sent
bool waterNotificationSent = false;    // Flag to track if a wet diaper notification has been sent

#define DHTPIN 4                       // Pin where the DHT sensor is connected
#define DHTTYPE DHT22                  // DHT 22 sensor type

DHT dht(DHTPIN, DHTTYPE);              // Initializing the DHT sensor for temperature and humidity

#define LED_PIN 27                     // Pin where the LED is connected

BlynkTimer timer;                      // Blynk timer to run functions at intervals
unsigned long lastRockingTime = 0;     // Tracks the last time the crib was rocked
const int rockingInterval = 2000;      // Interval for a full rocking cycle (in milliseconds)
bool rockingDirection = true;          // Direction of rocking (true = 180°, false = 0°)

// Variables for averaging sound level readings
const int numReadings = 10;            // Number of readings to average
int readings[numReadings];             // Array to store the readings
int readIndex = 0;                     // Index for the current reading
int total = 0;                         // Total of all readings
int average = 0;                       // The average value

// Function to send temperature data to Blynk app
void sendSensor() {
  float h = dht.readHumidity();
  float t = dht.readTemperature();     // Read temperature in Celsius

  if (isnan(h) || isnan(t)) {          // Check if the readings are valid
    Serial.println("Failed to read from DHT sensor!");
    return;
  }

  Blynk.virtualWrite(V0, t);           // Send temperature to virtual pin V0 on Blynk app
}

// Function to send water detection notification
void sendWaterNotification() {
  if (WiFi.status() == WL_CONNECTED) { // Check if WiFi is connected
    HTTPClient http;
    http.begin(waterServerName);       // Begin HTTP connection to the water notification webhook
    http.addHeader("Content-Type", "application/json");

    int httpResponseCode = http.POST("{\"message\":\"baby needs change\"}"); // Send POST request with a JSON payload
    Serial.print("HTTP Response code: ");
    Serial.println(httpResponseCode);

    if (httpResponseCode > 0) {        // Check if the request was successful
      String response = http.getString();
      Serial.println(response);
    } else {
      Serial.println("Error on sending POST");
    }
    http.end();                        // End the HTTP connection
  }
}

// Function to send a notification when the baby is crying
void sendNotification() {
  if (WiFi.status() == WL_CONNECTED) { // Check if WiFi is connected
    HTTPClient http;
    http.begin(serverName);            // Begin HTTP connection to the crying notification webhook
    http.addHeader("Content-Type", "application/json");

    int httpResponseCode = http.POST("{\"message\":\"baby crying\"}"); // Send POST request with a JSON payload
    Serial.print("HTTP Response code: ");
    Serial.println(httpResponseCode);

    if (httpResponseCode > 0) {        // Check if the request was successful
      String response = http.getString();
      Serial.println(response);
    } else {
      Serial.println("Error on sending POST");
    }
    http.end();                        // End the HTTP connection
  }
}

// Function to read sound levels and send to Blynk app
void sendVoiceLevel() {
  soundLevel = analogRead(soundSensorPin);    // Read the sound level from the sensor
  Serial.print("Voice Level: ");
  Serial.println(soundLevel);
  Blynk.virtualWrite(V4, soundLevel);         // Send sound level to virtual pin V4 on Blynk app
  delay(1000);
}

// Function to detect if the diaper is wet and send a notification
void checkSoilMoisture() {
  soilMoistureLevel = analogRead(soilSensorPin); // Read the moisture level from the sensor
  Serial.print("Soil Moisture Level: ");
  Serial.println(soilMoistureLevel);

  // Assuming a threshold value for soil moisture
  if (soilMoistureLevel < 2000) {          // Threshold for wet diaper
    if (!waterNotificationSent) {
      Serial.println("Diaper is wet. Sending notification...");
      sendWaterNotification();            // Send a notification if diaper is wet
      waterNotificationSent = true;       // Set the flag to indicate the notification has been sent
    }
    Blynk.virtualWrite(V3, 255);          // Turn on the LED in the Blynk app
  } else {
    if (waterNotificationSent) {
      Serial.println("Diaper is dry. Resetting notification flag...");
    }
    waterNotificationSent = false;        // Reset the flag when the diaper is dry
    Blynk.virtualWrite(V3, 0);            // Turn off the LED in the Blynk app
  }
}

void setup() {
  pinMode(soundSensorPin, INPUT);          // Set sound sensor pin as input
  pinMode(soilSensorPin, INPUT);           // Set soil moisture sensor pin as input
  pinMode(joystickButtonPin, INPUT_PULLUP); // Set joystick button pin as input with pull-up resistor
  pinMode(LED_PIN, OUTPUT);                // Set the LED pin as output
  digitalWrite(LED_PIN, 0);                // Ensure the LED is off at startup

  myServo.attach(23);                      // Attach the servo motor to pin 23
  Serial.begin(115200);                    // Begin serial communication at 115200 baud rate

  Serial.println("Mounting SPIFFS...");
  if (!SPIFFS.begin()) {                   // Mount SPIFFS (File system for storing MP3 files)
    Serial.println("Failed to mount file system");
    return;
  }

  out = new AudioOutputI2S();              // Set up I2S audio output
  out->SetGain(1);                         // Increase the gain to make the volume louder

  WiFi.begin(ssid, password);              // Begin WiFi connection with the given credentials
  unsigned long startAttemptTime = millis();

  // Non-blocking loop to connect to WiFi
  while (WiFi.status() != WL_CONNECTED && millis() - startAttemptTime < 20000) { // 20 seconds timeout
    Serial.print(".");
  }

  if (WiFi.status() == WL_CONNECTED) {     // Check if the connection was successful
    Serial.println();
    Serial.println("Connected to WiFi");
    Blynk.begin(auth, ssid, password);     // Begin Blynk service with the auth token
    dht.begin();                           // Begin DHT sensor

    // Setup functions to be called at regular intervals
    timer.setInterval(4000L, sendSensor);  // Send temperature data every 4 seconds
    timer.setInterval(7000L, sendVoiceLevel); // Send voice level data every 7 seconds
    timer.setInterval(5000L, checkSoilMoisture); // Check soil moisture level every 5 seconds
  } else {
    Serial.println();
    Serial.println("Failed to connect to WiFi");
  }

  // Initialize the readings array to zero
  for (int i = 0; i < numReadings; i++) {
    readings[i] = 0;
  }
}

// Function to control the rocking motion of the servo motor
void rockServo() {
  unsigned long currentTime = millis();  // Get the current time
  if (currentTime - lastRockingTime >= rockingInterval / 2) { // Check if it's time to change direction
    if (rockingDirection) {
      myServo.write(180);                // Move the servo to one end (180°)
    } else {
      myServo.write(0);                  // Move the servo to the other end (0°)
    }
    rockingDirection = !rockingDirection; // Toggle the direction for the next cycle
    lastRockingTime = currentTime;       // Update the last rocking time
  }
}

void loop() {
  if (WiFi.status() == WL_CONNECTED) {   // Check if WiFi is connected
    Blynk.run();                         // Run the Blynk service
    timer.run();                         // Run the Blynk timers

    // Check the joystick button state
    currentButtonState = digitalRead(joystickButtonPin);
    if (currentButtonState == LOW && lastButtonState == HIGH) { // Check if the button was pressed
      servoActive = !servoActive;        // Toggle the servo state
    }
    lastButtonState = currentButtonState;

    if (servoActive) {
      rockServo();                       // Rock the servo motor if active
    }

    // Update sound level readings for averaging
    total = total - readings[readIndex]; // Remove the oldest reading
    readings[readIndex] = analogRead(soundSensorPin); // Read the new sound level
    total = total + readings[readIndex]; // Add the new reading to the total
    readIndex = readIndex + 1;           // Advance to the next position in the array

    if (readIndex >= numReadings) {      // Wrap around to the beginning if needed
      readIndex = 0;
    }

    average = total / numReadings;       // Calculate the average sound level
    
    Serial.print("Average Voice Level: ");
    Serial.println(average);

    // Hysteresis: Prevents rapid toggling around the threshold
    int thresholdHigh = 2500;            // High threshold for crying detection
    int thresholdLow = 0;                // Low threshold for resetting the notification

    if (average > thresholdHigh) {       // If the sound level is high (crying detected)
      if (!notificationSent) {
        sendNotification();              // Send a crying notification
        notificationSent = true;         // Set the flag to indicate the notification has been sent
      }
    } else if (average < thresholdLow) {
      notificationSent = false;          // Reset the flag if the sound level is low
    }

    // Handle MP3 playback if active
    if (isPlaying) {
      if (mp3 && mp3->isRunning()) {
        if (!mp3->loop()) {              // Continue playing the MP3 file
          Serial.println("MP3 playback finished");
          mp3->stop();                   // Stop MP3 if it's finished playing
          isPlaying = false;
        }
      } else {
        Serial.println("MP3 is not running");
      }
    }
  }
}

// Blynk function to control LED based on button press in app
BLYNK_WRITE(V2) {
  int pinValue = param.asInt();          // Get the value of the virtual pin V2
  digitalWrite(LED_PIN, pinValue);       // Turn the LED on/off based on the button state
}

// Metadata callback function for ID3 tags in MP3 files
void MDCallback(void *cbData, const char *type, bool isUnicode, const char *string) {
  (void)cbData;
  Serial.printf("ID3 callback for: %s = '", type);
  if (isUnicode) {
    string += 2;                         // Adjust string pointer if data is Unicode
  }
  while (*string) {                      // Loop through the string
    char a = *(string++);
    if (isUnicode) {
      string++;                          // Increment pointer if Unicode
    }
    Serial.printf("%c", a);
  }
  Serial.printf("'\n");
  Serial.flush();
}

// Blynk function to control MP3 playback based on button press in app
BLYNK_WRITE(V1) {
  int pinValue = param.asInt();          // Get the value of the virtual pin V1
  Serial.print("Blynk V1 value: ");
  Serial.println(pinValue);
  if (pinValue == 1) {
    Serial.println("Switch ON: Starting MP3 playback");
    isPlaying = true;
    if (mp3 == nullptr) {
      file = new AudioFileSourceSPIFFS("/song1.mp3"); // Load the MP3 file from SPIFFS
      id3 = new AudioFileSourceID3(file); // Initialize ID3 for metadata extraction
      id3->RegisterMetadataCB(MDCallback, (void *)"ID3TAG"); // Register metadata callback
      mp3 = new AudioGeneratorMP3();     // Initialize MP3 generator
      mp3->begin(id3, out);              // Start MP3 playback
    }
  } else {
    Serial.println("Switch OFF: Stopping MP3 playback");
    isPlaying = false;
    if (mp3 && mp3->isRunning()) {
      mp3->stop();                       // Stop MP3 playback
      delete mp3;
      mp3 = nullptr;
      delete id3;
      id3 = nullptr;
      delete file;
      file = nullptr;
    }
  }
}
